import os


class TrainSettings:

    WEIGHTS_PATH: str = os.getenv("WEIGHTS_PATH", default="weights")


class KANSettings:

    KAN_IMPLEMENTATION: str = os.getenv(
        "KAN_IMPLEMENTATION", default="EFFICIENT_KAN"
    )

    def __init__(self) -> None:
        assert self.KAN_IMPLEMENTATION in ("EFFICIENT_KAN", "ORIGINAL_KAN")


class Settings:

    kan: KANSettings = KANSettings()
    train: TrainSettings = TrainSettings()


settings = Settings()



import os
from typing import List

import requests
from tqdm import tqdm

CHUNK_SIZE = 8192


def download_webtext(
    download_path: str = "datasets/webtext",
    splits: List[str] = ["train", "test", "valid"],
    base_url: str = "https://openaipublic.blob.core.windows.net/gpt-2/output-dataset/v1",  # noqa
):
    os.makedirs(download_path, exist_ok=True)
    for split in splits:
        file_path = f"{download_path}/webtext.{split}.jsonl"
        file_url = f"{base_url}/webtext.{split}.jsonl"

        # Check if file exists and get its size
        initial_pos = 0
        if os.path.exists(file_path):
            initial_pos = os.path.getsize(file_path)
            print(
                f"Resuming download of webtext.{split}.jsonl from {initial_pos} bytes"  # noqa
            )

        # Set up headers for resuming download
        headers = {"Range": f"bytes={initial_pos}-"}

        response = requests.get(file_url, stream=True, headers=headers)

        # If the server doesn't support range requests, start over
        if response.status_code == 416:
            print(
                f"Cannot resume download for webtext.{split}.jsonl. Starting from beginning."  # noqa
            )
            initial_pos = 0
            headers = {}
            response = requests.get(file_url, stream=True)

        response.raise_for_status()  # Raise HTTP errors
        total_size = (
            int(response.headers.get("content-length", 0)) + initial_pos
        )

        # Open the local file for writing in binary mode, appending if resuming
        mode = "ab" if initial_pos > 0 else "wb"
        with open(file_path, mode) as file, tqdm(
            desc=f"Downloading webtext.{split}.jsonl",
            initial=initial_pos,
            total=total_size,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as progress_bar:
            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                size = file.write(chunk)
                progress_bar.update(size)

        # Verify file size after download
        if os.path.getsize(file_path) != total_size:
            print(
                f"Warning: Downloaded file size does not match expected size for webtext.{split}.jsonl"  # noqa
            )


def download_tinyshakespeare(
    download_path: str = "datasets/tinyshakespeare",
    base_url: str = "https://raw.githubusercontent.com/karpathy/char-rnn/master/data/tinyshakespeare",  # noqa
):
    os.makedirs(download_path, exist_ok=True)

    response = requests.get(f"{base_url}/input.txt", stream=True)
    response.raise_for_status()  # Raise HTTP errors
    total_size = int(response.headers.get("content-length", 0))

    # Open a local file for writing in binary mode
    with open(f"{download_path}/input.txt", "wb") as file, tqdm(
        desc=f"Downloading {download_path}/input.txt",
        total=total_size,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as progress_bar:
        for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
            size = file.write(chunk)
            progress_bar.update(size)


def download_mnist():
    from .dataset import MNISTDataset

    MNISTDataset("train", "gpt2")
    MNISTDataset("test", "gpt2")


def main(args):
    if args.dataset == "webtext":
        download_webtext()
    elif args.dataset == "tinyshakespeare":
        download_tinyshakespeare()
    elif args.dataset == "mnist":
        download_mnist()
    else:
        raise NotImplementedError


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser("KAN-GPT Trainer")
    parser.add_argument(
        "--dataset",
        choices=["webtext", "tinyshakespeare", "mnist"],
        default="tinyshakespeare",
    )

    args = parser.parse_args()

    main(args)



import random
from tempfile import TemporaryDirectory

import torch
from kan_gpt.mingpt.model import GPT as MLP_GPT
from kan_gpt.mingpt.utils import set_seed, setup_logging, CfgNode as CN

VOCAB_SIZE = 8
BLOCK_SIZE = 16
MODEL_TYPE = "gpt-pico"


def get_gpt_model() -> MLP_GPT:
    model_config = MLP_GPT.get_default_config()
    model_config.model_type = MODEL_TYPE
    model_config.vocab_size = VOCAB_SIZE
    model_config.block_size = BLOCK_SIZE
    model = MLP_GPT(model_config)
    return model


def test_forward():
    with torch.no_grad():
        model = get_gpt_model()
        x = torch.zeros((1, BLOCK_SIZE), dtype=torch.long)

        y, loss = model.forward(x)

        assert y.shape == (
            1,
            BLOCK_SIZE,
            VOCAB_SIZE,
        ), f"Shape mismatch: {y.shape}"


def test_backward():
    model = get_gpt_model()
    x = torch.zeros((1, BLOCK_SIZE), dtype=torch.long)

    # Make sure grads exist
    requires_grad_set = set()
    for param in model.parameters():
        if param.requires_grad:
            requires_grad_set.add(param)
    assert len(requires_grad_set) > 0, "requires_grad is not set"

    y, loss = model.forward(x)

    assert y.shape == (1, BLOCK_SIZE, VOCAB_SIZE), f"Shape mismatch: {y.shape}"

    loss = y.mean()
    loss.backward()

    # Make sure grads exist
    grad_set = set()
    for param in model.parameters():
        if isinstance(param.grad, torch.Tensor):
            grad_set.add(param)
    assert len(grad_set) > 0, f"Tensor.grad missing"


def test_forward_batched():
    with torch.no_grad():
        model = get_gpt_model()
        x = torch.zeros((2, BLOCK_SIZE), dtype=torch.long)

        y, loss = model.forward(x)

        assert y.shape == (
            2,
            BLOCK_SIZE,
            VOCAB_SIZE,
        ), f"Shape mismatch: {y.shape}"


def test_backward_batched():
    model = get_gpt_model()
    x = torch.zeros((2, BLOCK_SIZE), dtype=torch.long)

    # Make sure grads exist
    requires_grad_set = set()
    for param in model.parameters():
        if param.requires_grad:
            requires_grad_set.add(param)
    assert len(requires_grad_set) > 0, "requires_grad is not set"

    y, loss = model.forward(x)

    assert y.shape == (2, BLOCK_SIZE, VOCAB_SIZE), f"Shape mismatch: {y.shape}"

    loss = y.mean()
    loss.backward()

    # Make sure grads exist
    grad_set = set()
    for param in model.parameters():
        if isinstance(param.grad, torch.Tensor):
            grad_set.add(param)
    assert len(grad_set) > 0, f"Tensor.grad missing"


def test_CN():
    C = CN()
    C.device = "auto"
    assert C.device == "auto", "Unable to set param"


def test_seed_set():
    seed = 0
    set_seed(seed)

    rr1 = random.random()
    rr2 = random.random()

    set_seed(seed)

    rr3 = random.random()

    assert rr1 == rr3, "seed not set correctly"
    assert rr1 != rr2, "seed not set correctly"


def test_setup_logging():
    C = CN()
    with TemporaryDirectory() as folder:
        C.system = CN()
        C.system.work_dir = folder
        setup_logging(C)



from kan_gpt.prompt import main

VOCAB_SIZE = 8
BLOCK_SIZE = 16
MODEL_TYPE = "gpt-pico"


def test_train():
    class Args:
        model_type = MODEL_TYPE
        model_path = None
        max_tokens = 3
        prompt = "Bangalore is often described as the "
        architecture = "KAN"
        device = "cpu"

    args = Args()
    main(args)



"""
bpe is short for Byte Pair Encoder. It translates arbitrary utf-8 strings into
sequences of integers, where each integer represents small chunks of commonly
occuring characters. This implementation is based on openai's gpt2 encoder.py:
https://github.com/openai/gpt-2/blob/master/src/encoder.py
but was mildly modified because the original implementation is a bit confusing.
I also tried to add as many comments as possible, my own understanding of what's
going on.
"""

import json
import os
import re

import requests
import torch

# -----------------------------------------------------------------------------


def bytes_to_unicode():
    """
    Every possible byte (really an integer 0..255) gets mapped by OpenAI to a unicode
    character that represents it visually. Some bytes have their appearance preserved
    because they don't cause any trouble. These are defined in list bs. For example:
    chr(33) returns "!", so in the returned dictionary we simply have d[33] -> "!".
    However, chr(0), for example, is '\x00', which looks ugly. So OpenAI maps these
    bytes, into new characters in a range where chr() returns a single nice character.
    So in the final dictionary we have d[0] -> 'Ā' instead, which is just chr(0 + 2**8).
    In particular, the space character is 32, which we can see by ord(' '). Instead,
    this function will shift space (32) by 256 to 288, so d[32] -> 'Ġ'.
    So this is just a simple one-to-one mapping of bytes 0..255 into unicode characters
    that "look nice", either in their original form, or a funny shifted character
    like 'Ā', or 'Ġ', etc.
    """
    # the 188 integers that render fine in their original form and need no shifting
    bs = (
        list(range(ord("!"), ord("~") + 1))
        + list(range(ord("¡"), ord("¬") + 1))
        + list(range(ord("®"), ord("ÿ") + 1))
    )
    cs = bs[
        :
    ]  # all integers b in bs will simply map to chr(b) in the output dict
    # now get the representations of the other 68 integers that do need shifting
    # each will get mapped chr(256 + n), where n will grow from 0...67 in the loop
    n = 0
    for b in range(2**8):
        if b not in bs:
            # if this byte is "ugly" then map it to the next available "nice" character
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    d = dict(zip(bs, cs))
    return d


def get_pairs(word):
    """
    Return all bigrams as a set of tuples, of consecutive elements in the iterable word.
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


class Encoder:

    def __init__(self, encoder, bpe_merges):
        # byte encoder/decoder
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        # bpe token encoder/decoder
        self.encoder = encoder
        self.decoder = {v: k for k, v in self.encoder.items()}
        # bpe merge list that defines the bpe "tree", of tuples (a,b) that are to merge to token ab
        self.bpe_ranks = dict(zip(bpe_merges, range(len(bpe_merges))))
        # the splitting pattern used for pre-tokenization
        # Should haved added re.IGNORECASE so BPE merges can happen for capitalized versions of contractions <-- original openai comment
        """
        ok so what is this regex looking for, exactly?
        python re reference: https://docs.python.org/3/library/re.html
        - the vertical bars | is OR, so re.findall will chunkate text as the pieces match, from left to right
        - '\'s' would split up things like Andrej's -> (Andrej, 's)
        - ' ?\p{L}': optional space followed by 1+ unicode code points in the category "letter"
        - ' ?\p{N}': optional space followed by 1+ unicode code points in the category "number"
        - ' ?[^\s\p{L}\p{N}]+': optional space, then 1+ things that are NOT a whitespace, letter or number
        - '\s+(?!\S)': 1+ whitespace characters (e.g. space or tab or etc) UNLESS they are followed by non-whitespace
                       so this will consume whitespace characters in a sequence but exclude the last whitespace in
                       that sequence. that last whitespace has the opportunity to then match the optional ' ?' in
                       earlier patterns.
        - '\s+': 1+ whitespace characters, intended probably to catch a full trailing sequence of whitespaces at end of string
        So TLDR:
        - we are special casing a few common apostrophe constructs ('s, 't, 're, ...) and making those into separate tokens
        - we then separate out strings into consecutive chunks of 1) letters, 2) numbers, 3) non-letter-numbers, 4) whitespaces
        """
        self.pat = re.compile(
            r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+"""
        )
        self.cache = {}

    def bpe(self, token):
        """
        this function uses self.bpe_ranks to iteratively merge all the possible bpe tokens
        up the tree. token is a string of one individual 'word' (after regex tokenization)
        and after byte encoding, e.g. 'Ġthere'.
        """
        # token is a string of one individual 'word', after byte encoding, e.g. 'Ġthere'

        # memoization, for efficiency
        if token in self.cache:
            return self.cache[token]

        word = tuple(
            token
        )  # individual characters that make up the token, in a tuple
        pairs = get_pairs(word)  # get all bigrams

        if not pairs:
            return token

        while True:

            # find the next lowest rank bigram that can be merged
            bigram = min(
                pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf"))
            )
            if bigram not in self.bpe_ranks:
                break  # no more bigrams are eligible to be merged
            first, second = bigram

            # we will now replace all occurences of (first, second) in the list of current
            # words into one merged token first_second, in the output list new_words
            new_word = []
            i = 0
            while i < len(word):

                # find the next occurence of first in the sequence of current words
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:  # noqa
                    new_word.extend(word[i:])
                    break

                # if this occurence is also followed by second, then merge them into one
                if (
                    word[i] == first
                    and i < len(word) - 1
                    and word[i + 1] == second
                ):
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1

            # all occurences of (first, second) have been merged to first_second
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)

        # concat all words into a string, and use ' ' as the separator. Note that
        # by now all characters have been byte encoded, guaranteeing that ' ' is
        # not used in the actual data and is a 'special' delimiter character
        word = " ".join(word)

        # cache the result and return
        self.cache[token] = word
        return word

    def encode(self, text):
        """string goes in, list of integers comes out"""
        bpe_idx = []
        # pre-tokenize the input text into string tokens (words, roughly speaking)
        tokens = re.findall(self.pat, text)
        # process each token into BPE integers
        for token in tokens:
            # encode the token as a bytes (b'') object
            token_bytes = token.encode("utf-8")
            # translate all bytes to their unicode string representation and flatten
            token_translated = "".join(
                self.byte_encoder[b] for b in token_bytes
            )
            # perform all the applicable bpe merges according to self.bpe_ranks
            token_merged = self.bpe(token_translated).split(" ")
            # translate all bpe tokens to integers
            token_ix = [self.encoder[bpe_token] for bpe_token in token_merged]
            # extend our running list of all output integers
            bpe_idx.extend(token_ix)
        return bpe_idx

    def encode_and_show_work(self, text):
        """debugging function, same as encode but returns all intermediate work"""
        bpe_idx = []
        parts = []
        tokens = re.findall(self.pat, text)
        for token in tokens:
            token_bytes = token.encode("utf-8")
            token_translated = "".join(
                self.byte_encoder[b] for b in token_bytes
            )
            token_merged = self.bpe(token_translated).split(" ")
            token_ix = [self.encoder[bpe_token] for bpe_token in token_merged]
            bpe_idx.extend(token_ix)
            parts.append(
                {
                    "token": token,
                    "token_bytes": token_bytes,
                    "token_translated": token_translated,
                    "token_merged": token_merged,
                    "token_ix": token_ix,
                }
            )
        out = {
            "bpe_idx": bpe_idx,  # the actual output sequence
            "tokens": tokens,  # result of pre-tokenization
            "parts": parts,  # intermediates for each token part
        }
        return out

    def decode(self, bpe_idx):
        """list of integers comes in, string comes out"""
        # inverse map the integers to get the tokens
        tokens_merged = [self.decoder[token] for token in bpe_idx]
        # inverse the byte encoder, e.g. recovering 'Ġ' -> ' ', and get the bytes
        tokens_flat = "".join(tokens_merged)
        tokens_bytes = bytearray([self.byte_decoder[c] for c in tokens_flat])
        # recover the full utf-8 string
        text = tokens_bytes.decode("utf-8", errors="replace")
        return text


def get_file(local_file, remote_file):
    """downloads remote_file to local_file if necessary"""
    if not os.path.isfile(local_file):
        print(f"downloading {remote_file} to {local_file}")
        response = requests.get(remote_file)
        open(local_file, "wb").write(response.content)


def get_encoder():
    """
    Returns an instance of the GPT BPE Encoder/Decoder
    and handles caching of "database" files.
    """
    home_dir = os.path.expanduser("~")
    cache_dir = os.path.join(home_dir, ".cache", "mingpt")
    os.makedirs(cache_dir, exist_ok=True)

    # load encoder.json that has the raw mappings from token -> bpe index
    encoder_local_file = os.path.join(cache_dir, "encoder.json")
    encoder_remote_file = "https://openaipublic.blob.core.windows.net/gpt-2/models/124M/encoder.json"
    get_file(encoder_local_file, encoder_remote_file)
    with open(encoder_local_file, "r") as f:
        encoder = json.load(f)
    assert (
        len(encoder) == 50257
    )  # 256 individual byte tokens, 50,000 merged tokens, and 1 special <|endoftext|> token

    # load vocab.bpe that contains the bpe merges, i.e. the bpe tree structure
    # in the form tuples (a, b), that indicate that (a, b) is to be merged to one token ab
    vocab_local_file = os.path.join(cache_dir, "vocab.bpe")
    vocab_remote_file = "https://openaipublic.blob.core.windows.net/gpt-2/models/124M/vocab.bpe"
    get_file(vocab_local_file, vocab_remote_file)
    with open(vocab_local_file, "r", encoding="utf-8") as f:
        bpe_data = f.read()
    # light postprocessing: strip the version on first line and the last line is a blank
    bpe_merges = [
        tuple(merge_str.split()) for merge_str in bpe_data.split("\n")[1:-1]
    ]
    assert len(bpe_merges) == 50000  # 50,000 merged tokens

    # construct the Encoder object and return
    enc = Encoder(encoder, bpe_merges)
    return enc


# -----------------------------------------------------------------------------


class BPETokenizer:
    """PyTorch-aware class that wraps the Encoder above"""

    def __init__(self):
        self.encoder = get_encoder()

    def __call__(self, text, return_tensors="pt"):
        # PyTorch only; here because we want to match huggingface/transformers interface
        assert return_tensors == "pt"
        # single string input for now, in the future potentially a list of strings
        assert isinstance(text, str)
        # encode and create a "batch dimension" of 1
        idx = [self.encoder.encode(text)]
        # wrap into PyTorch tensor
        out = torch.tensor(idx, dtype=torch.long)
        return out

    def decode(self, idx):
        # ensure a simple 1D tensor for now
        assert idx.ndim == 1
        # decode indices to text
        text = self.encoder.decode(idx.tolist())
        return text


if __name__ == "__main__":

    # here is an encoding example
    text = "Hello!! I'm Andrej Karpathy. It's 2022. w00t :D 🤗"
    e = get_encoder()
    r = e.encode_and_show_work(text)

    print("Original text is:")
    print(text)
    print(
        "First the text gets pre-tokenized, broken up into chunks, the outcome is:"
    )
    print(r["tokens"])
    # ['Hello', '!!', ' I', "'m", ' Andrej', ' Karpathy', '.', ' It', "'s", ' 2022', '.', ' w', '00', 't', ' :', 'D', ' 🤗']
    print("Then we iterate over each chunk and process them in turn...")
    for part in r["parts"]:
        print(part)
    # {'token': 'Hello', 'token_bytes': b'Hello', 'token_translated': 'Hello', 'token_merged': ['Hello'], 'token_ix': [15496]}
    # {'token': '!!', 'token_bytes': b'!!', 'token_translated': '!!', 'token_merged': ['!!'], 'token_ix': [3228]}
    # {'token': ' I', 'token_bytes': b' I', 'token_translated': 'ĠI', 'token_merged': ['ĠI'], 'token_ix': [314]}
    # {'token': "'m", 'token_bytes': b"'m", 'token_translated': "'m", 'token_merged': ["'m"], 'token_ix': [1101]}
    # {'token': ' Andrej', 'token_bytes': b' Andrej', 'token_translated': 'ĠAndrej', 'token_merged': ['ĠAndre', 'j'], 'token_ix': [10948, 73]}
    # {'token': ' Karpathy', 'token_bytes': b' Karpathy', 'token_translated': 'ĠKarpathy', 'token_merged': ['ĠK', 'arp', 'athy'], 'token_ix': [509, 5117, 10036]}
    # {'token': '.', 'token_bytes': b'.', 'token_translated': '.', 'token_merged': ['.'], 'token_ix': [13]}
    # {'token': ' It', 'token_bytes': b' It', 'token_translated': 'ĠIt', 'token_merged': ['ĠIt'], 'token_ix': [632]}
    # {'token': "'s", 'token_bytes': b"'s", 'token_translated': "'s", 'token_merged': ["'s"], 'token_ix': [338]}
    # {'token': ' 2022', 'token_bytes': b' 2022', 'token_translated': 'Ġ2022', 'token_merged': ['Ġ2022'], 'token_ix': [33160]}
    # {'token': '.', 'token_bytes': b'.', 'token_translated': '.', 'token_merged': ['.'], 'token_ix': [13]}
    # {'token': ' w', 'token_bytes': b' w', 'token_translated': 'Ġw', 'token_merged': ['Ġw'], 'token_ix': [266]}
    # {'token': '00', 'token_bytes': b'00', 'token_translated': '00', 'token_merged': ['00'], 'token_ix': [405]}
    # {'token': 't', 'token_bytes': b't', 'token_translated': 't', 'token_merged': ['t'], 'token_ix': [83]}
    # {'token': ' :', 'token_bytes': b' :', 'token_translated': 'Ġ:', 'token_merged': ['Ġ:'], 'token_ix': [1058]}
    # {'token': 'D', 'token_bytes': b'D', 'token_translated': 'D', 'token_merged': ['D'], 'token_ix': [35]}
    # {'token': ' 🤗', 'token_bytes': b' \xf0\x9f\xa4\x97', 'token_translated': 'ĠðŁ¤Ĺ', 'token_merged': ['ĠðŁ', '¤', 'Ĺ'], 'token_ix': [12520, 97, 245]}
    # (refer to the code inside Encoder.encode for what these intermediates are)
    print(
        "and the final outcome is concatenating and flattening all the token_ix:"
    )
    print(r["bpe_idx"])
    # [15496, 3228, 314, 1101, 10948, 73, 509, 5117, 10036, 13, 632, 338, 33160, 13, 266, 405, 83, 1058, 35, 12520, 97, 245]
    # this would then become the integer input sequence to the transformer
    print("ready to feed into a Transformer!")



import torch
import torch.nn.functional as F


class KANLinear(torch.nn.Module):
    def __init__(
        self,
        in_features,
        out_features,
        grid_size=5,
        spline_order=3,
        scale_noise=0.1,
        scale_base=1.0,
        scale_spline=1.0,
        enable_standalone_scale_spline=True,
        base_activation=torch.nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
    ):
        super(KANLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.grid_size = grid_size
        self.spline_order = spline_order

        h = (grid_range[1] - grid_range[0]) / grid_size
        grid = (
            (
                torch.arange(-spline_order, grid_size + spline_order + 1) * h
                + grid_range[0]
            )
            .expand(in_features, -1)
            .contiguous()
        )
        self.register_buffer("grid", grid)

        self.base_weight = torch.nn.Parameter(
            torch.Tensor(out_features, in_features)
        )
        self.spline_weight = torch.nn.Parameter(
            torch.Tensor(out_features, in_features, grid_size + spline_order)
        )
        if enable_standalone_scale_spline:
            self.spline_scaler = torch.nn.Parameter(
                torch.Tensor(out_features, in_features)
            )

        self.scale_noise = scale_noise
        self.scale_base = scale_base
        self.scale_spline = scale_spline
        self.enable_standalone_scale_spline = enable_standalone_scale_spline
        self.base_activation = base_activation()
        self.grid_eps = grid_eps

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.base_weight, gain=self.scale_base)
        with torch.no_grad():
            noise = (
                (
                    torch.rand(
                        self.grid_size + 1, self.in_features, self.out_features
                    )
                    - 1 / 2
                )
                * self.scale_noise
                / self.grid_size
            )
            self.spline_weight.data.copy_(
                (
                    self.scale_spline
                    if not self.enable_standalone_scale_spline
                    else 1.0
                )
                * self.curve2coeff(
                    self.grid.T[self.spline_order : -self.spline_order],
                    noise,
                )
            )
            if self.enable_standalone_scale_spline:
                torch.nn.init.constant_(self.spline_scaler, self.scale_spline)

    def b_splines(self, x: torch.Tensor):
        """
        Compute the B-spline bases for the given input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).

        Returns:
            torch.Tensor: B-spline bases tensor of shape \
                (batch_size, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2 and x.size(1) == self.in_features

        grid: torch.Tensor = (
            self.grid  # type: ignore
        )  # (in_features, grid_size + 2 * spline_order + 1)
        x = x.unsqueeze(-1)
        bases = ((x >= grid[:, :-1]) & (x < grid[:, 1:])).to(x.dtype)
        for k in range(1, self.spline_order + 1):
            bases = (
                (x - grid[:, : -(k + 1)])
                / (grid[:, k:-1] - grid[:, : -(k + 1)])
                * bases[:, :, :-1]
            ) + (
                (grid[:, k + 1 :] - x)
                / (grid[:, k + 1 :] - grid[:, 1:(-k)])
                * bases[:, :, 1:]
            )

        assert bases.size() == (
            x.size(0),
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return bases.contiguous()

    def curve2coeff(self, x: torch.Tensor, y: torch.Tensor):
        """
        Compute the coefficients of the curve that interpolates the given
        points.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).
            y (torch.Tensor): Output tensor of shape \
                (batch_size, in_features, out_features).

        Returns:
            torch.Tensor: Coefficients tensor of shape \
                (out_features, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2 and x.size(1) == self.in_features
        assert y.size() == (x.size(0), self.in_features, self.out_features)

        A = self.b_splines(x).transpose(
            0, 1
        )  # (in_features, batch_size, grid_size + spline_order)
        B = y.transpose(0, 1)  # (in_features, batch_size, out_features)
        solution = torch.linalg.lstsq(
            A, B
        ).solution  # (in_features, grid_size + spline_order, out_features)
        result = solution.permute(
            2, 0, 1
        )  # (out_features, in_features, grid_size + spline_order)

        assert result.size() == (
            self.out_features,
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return result.contiguous()

    @property
    def scaled_spline_weight(self):
        return self.spline_weight * (
            self.spline_scaler.unsqueeze(-1)
            if self.enable_standalone_scale_spline
            else 1.0
        )

    def forward(self, x: torch.Tensor):
        assert x.dim() == 2 and x.size(1) == self.in_features

        base_output = F.linear(self.base_activation(x), self.base_weight)
        spline_output = F.linear(
            self.b_splines(x).view(x.size(0), -1),
            self.scaled_spline_weight.view(self.out_features, -1),
        )
        return base_output + spline_output

    @torch.no_grad()
    def update_grid(self, x: torch.Tensor, margin=0.01):
        assert x.dim() == 2 and x.size(1) == self.in_features
        batch = x.size(0)

        splines = self.b_splines(x)  # (batch, in, coeff)
        splines = splines.permute(1, 0, 2)  # (in, batch, coeff)
        orig_coeff = self.scaled_spline_weight  # (out, in, coeff)
        orig_coeff = orig_coeff.permute(1, 2, 0)  # (in, coeff, out)
        unreduced_spline_output = torch.bmm(
            splines, orig_coeff
        )  # (in, batch, out)
        unreduced_spline_output = unreduced_spline_output.permute(
            1, 0, 2
        )  # (batch, in, out)

        # sort each channel individually to collect data distribution
        x_sorted = torch.sort(x, dim=0)[0]
        grid_adaptive = x_sorted[
            torch.linspace(
                0,
                batch - 1,
                self.grid_size + 1,
                dtype=torch.int64,
                device=x.device,
            )
        ]

        uniform_step = (
            x_sorted[-1] - x_sorted[0] + 2 * margin
        ) / self.grid_size
        grid_uniform = (
            torch.arange(
                self.grid_size + 1, dtype=torch.float32, device=x.device
            ).unsqueeze(1)
            * uniform_step
            + x_sorted[0]
            - margin
        )

        grid = (
            self.grid_eps * grid_uniform + (1 - self.grid_eps) * grid_adaptive
        )
        grid = torch.concatenate(
            [
                grid[:1]
                - uniform_step
                * torch.arange(
                    self.spline_order, 0, -1, device=x.device
                ).unsqueeze(1),
                grid,
                grid[-1:]
                + uniform_step
                * torch.arange(
                    1, self.spline_order + 1, device=x.device
                ).unsqueeze(1),
            ],
            dim=0,
        )

        self.grid.copy_(grid.T)  # type: ignore
        self.spline_weight.data.copy_(
            self.curve2coeff(x, unreduced_spline_output)
        )

    def regularization_loss(
        self, regularize_activation=1.0, regularize_entropy=1.0
    ):
        """
        Compute the regularization loss.

        This is a dumb simulation of the original L1 regularization as
        stated in the paper, since the original one requires computing
        absolutes and entropy from the expanded
        (batch, in_features, out_features) intermediate tensor, which is
        hidden behind the F.linear function if we want an memory
        efficient implementation.

        The L1 regularization is now computed as mean absolute value of the
        spline weights. The authors implementation also includes this term
        in addition to the sample-based regularization.
        """
        l1_fake = self.spline_weight.abs().mean(-1)
        regularization_loss_activation = l1_fake.sum()
        p = l1_fake / regularization_loss_activation
        regularization_loss_entropy = -torch.sum(p * p.log())
        return (
            regularize_activation * regularization_loss_activation
            + regularize_entropy * regularization_loss_entropy
        )


class KAN(torch.nn.Module):
    def __init__(
        self,
        width,
        grid=3,
        k=3,
        noise_scale=0.1,
        noise_scale_base=1.0,
        scale_spline=1.0,
        base_fun=torch.nn.SiLU,
        grid_eps=0.02,
        grid_range=[-1, 1],
        bias_trainable=True,
    ):
        super(KAN, self).__init__()
        self.grid_size = grid
        self.spline_order = k
        self.bias_trainable = bias_trainable  # TODO

        self.layers = torch.nn.ModuleList()
        for in_features, out_features in zip(width, width[1:]):
            self.layers.append(
                KANLinear(
                    in_features,
                    out_features,
                    grid_size=grid,
                    spline_order=grid,
                    scale_noise=noise_scale,
                    scale_base=noise_scale_base,
                    scale_spline=scale_spline,
                    base_activation=base_fun,
                    grid_eps=grid_eps,
                    grid_range=grid_range,
                )
            )

    def forward(self, x: torch.Tensor, update_grid=False):
        B, C, T = x.shape

        x = x.view(-1, T)

        for layer in self.layers:
            if update_grid:
                layer.update_grid(x)
            x = layer(x)

        U = x.shape[1]

        x = x.view(B, C, U)

        return x

    def regularization_loss(
        self, regularize_activation=1.0, regularize_entropy=1.0
    ):
        return sum(
            layer.regularization_loss(
                regularize_activation, regularize_entropy
            )
            for layer in self.layers
        )



import torch

import wandb
from kan_gpt.train import main


def wandb_sweep():
    run = wandb.init(resume="allow", anonymous="must")

    class Args:
        model_type = wandb.config.model_type
        dummy_dataset = wandb.config.dummy_dataset
        learning_rate = wandb.config.learning_rate
        max_iters = wandb.config.max_iters
        num_workers = wandb.config.num_workers
        batch_size = wandb.config.batch_size
        dataset = wandb.config.dataset
        architecture = wandb.config.architecture
        device = wandb.config.device

    run_args = Args()

    if "cuda" in run_args.device:
        torch.cuda.empty_cache()

    main(args=run_args, run=run)

    if "cuda" in run_args.device:
        torch.cuda.empty_cache()


def sweep(args):
    sweep_configuration = {
        "method": "random",
        "name": "sweep",
        "metric": {"goal": "minimize", "name": "test_loss"},
        "parameters": {
            "model_type": {"values": args.model_type},
            "batch_size": {"values": args.batch_size},
            "dummy_dataset": {"values": args.dummy_dataset},
            "learning_rate": {"values": args.learning_rate},
            "max_iters": {"values": args.max_iters},
            "num_workers": {"values": args.num_workers},
            "dataset": {"values": args.dataset},
            "architecture": {"values": args.architecture},
            "device": {"values": args.device},
        },
    }

    sweep_id = wandb.sweep(sweep_configuration, project="KAN-GPT")
    print("sweep_id (generated)", sweep_id)

    wandb.agent(sweep_id, function=wandb_sweep)


if __name__ == "__main__":

    class SweepArgs:
        model_type = ["gpt-mini", "gpt-micro", "gpt-nano", "gpt-pico"]
        dummy_dataset = [
            False,
        ]
        learning_rate = [5e-5, 5e-6, 5e-7]
        max_iters = [
            8000,
        ]
        num_workers = [
            0,
        ]
        batch_size = [1, 2, 3, 4]
        dataset = [
            "tinyshakespeare",
        ]
        architecture = ["MLP", "KAN"]
        device = [
            "cuda",
        ]

    sweep(SweepArgs())



import os
from tempfile import TemporaryDirectory
import pytest
import requests_mock

from kan_gpt.download_dataset import download_tinyshakespeare, download_webtext


@pytest.fixture
def mock_api():
    with requests_mock.Mocker() as m:
        yield m


def test_download_tinyshakespeare(mock_api):
    import requests

    base_url = "http://tinyshakespeare/v1"
    response_data = "Out of thy sleep. What is it thou didst say?"

    with TemporaryDirectory() as download_path:
        mock_api.get(f"{base_url}/input.txt", text=response_data)

        assert response_data == requests.get(f"{base_url}/input.txt").text

        download_tinyshakespeare(
            download_path=download_path, base_url=base_url
        )

        file_path = os.path.join(download_path, "input.txt")
        assert os.path.exists(file_path), f"File not found: {file_path}"


def test_download_webtext(mock_api):

    splits = ["test", "train", "valid"]
    base_url = "http://webtext/v1"
    response_data = '{"id": 259999, "ended": true, "length": 488, "text": "This post goes out to those of you who have not yet experienced the pure joy of brewing your morning joe with a Keurig. To those who have wondered just how those little K-cup thingys work and especially to those who have not heard of this new fangled technology at all (you must be living under a rock if you fall in that category!). Keurig brewers are specially designed to extract coffee from a K-cup. A K-cup is a highly engineered, technologically sophisticated mini coffee brewer inside of a very tiny body. K-cups consist of four parts:\n\nThe outer plastic casing. This is a special designed housing that blocks out moisture, light and air allowing your precious coffee to remain perfectly fresh until you are ready to brew it.\n\nA permeable paper filter. The paper filter allows for optimal flavor extraction.\n\nA foil seal. The foil seal keep the coffee air tight and blocks out oxygen and humidity.\n\nK-cups contain a perfectly measured amount of coffee creating a consistent brew cup after cup.\n\nTo brew coffee with a Keurig, the first step is to place the K-cup in the brew chamber of the unit. When the handle is pulled down closing the chamber, a small needle will puncture the foil lid of the K-cup penetrating the coffee with pressurized water at high temperature.The coffee is filtered through the paper. At the same time a second needle punctures the bottom of the K-cup, allowing freshly brewed coffee to pour into your mug.\n\nThis video does a nice job of demonstrating the brewing process:\n\nKeurig K-cup technology is incredibly revolutionary yet incredibly simple at the same time. The creators of the Keurig system shrunk coffee brewing down to its smallest form- one cup at a time. The K-cup is really just a tiny little drip style coffee brewer. The concept is exactly the same however, the technology of the Keurig itself is far more sophisticated.\n\nMany have tried to duplicate this technology but few have come close to matching the quality. So many consumers choose to brew their own coffee grinds with special adapters. Although they seem like a cost saver up front, we have yet to find one that operates as easily and tastes as good as a genuine Kcup.\n\nIf you are reading this article you may also be interested in:"}'  # noqa

    with TemporaryDirectory() as download_path:
        for split in splits:
            mock_api.get(
                f"{base_url}/webtext.{split}.jsonl", text=response_data
            )

        download_webtext(
            download_path=download_path,
            base_url=base_url,
            splits=splits,
        )

        for split in splits:
            file_path = f"{download_path}/webtext.{split}.jsonl"
            assert os.path.exists(file_path), f"File not found: {file_path}"



import torch
import torch.nn as nn

from .utils import SYMBOLIC_LIB, fit_params


class Symbolic_KANLayer(nn.Module):
    """
    KANLayer class

    Attributes:
    -----------
        in_dim: int
            input dimension
        out_dim: int
            output dimension
        funs: 2D array of torch functions (or lambda functions)
            symbolic functions (torch)
        funs_name: 2D arry of str
            names of symbolic functions
        funs_sympy: 2D array of sympy functions (or lambda functions)
            symbolic functions (sympy)
        affine: 3D array of floats
            affine transformations of inputs and outputs

    Methods:
    --------
        __init__():
            initialize a Symbolic_KANLayer
        forward():
            forward
        get_subset():
            get subset of the KANLayer (used for pruning)
        fix_symbolic():
            fix an activation function to be symbolic
    """

    def __init__(self, in_dim=3, out_dim=2):
        """
        initialize a Symbolic_KANLayer (activation functions are initialized to be identity functions)

        Args:
        -----
            in_dim : int
                input dimension
            out_dim : int
                output dimension

        Returns:
        --------
            self

        Example
        -------
        >>> sb = Symbolic_KANLayer(in_dim=3, out_dim=3)
        >>> len(sb.funs), len(sb.funs[0])
        (3, 3)
        """
        super(Symbolic_KANLayer, self).__init__()
        self.out_dim = out_dim
        self.in_dim = in_dim
        self.mask = torch.nn.Parameter(
            torch.zeros(out_dim, in_dim)
        ).requires_grad_(False)
        # torch
        self.funs = [
            [lambda x: x for i in range(self.in_dim)]
            for j in range(self.out_dim)
        ]
        # name
        self.funs_name = [
            ["" for i in range(self.in_dim)] for j in range(self.out_dim)
        ]
        # sympy
        self.funs_sympy = [
            ["" for i in range(self.in_dim)] for j in range(self.out_dim)
        ]

        self.affine = torch.nn.Parameter(torch.zeros(out_dim, in_dim, 4))
        # c*f(a*x+b)+d

    def forward(self, x):
        """
        forward

        Args:
        -----
            x : 2D array
                inputs, shape (batch, input dimension)

        Returns:
        --------
            y : 2D array
                outputs, shape (batch, output dimension)
            postacts : 3D array
                activations after activation functions but before summing on nodes

        Example
        -------
        >>> sb = Symbolic_KANLayer(in_dim=3, out_dim=5)
        >>> x = torch.normal(0,1,size=(100,3))
        >>> y, postacts = sb(x)
        >>> y.shape, postacts.shape
        (torch.Size([100, 5]), torch.Size([100, 5, 3]))
        """

        batch = x.shape[0]  # noqa
        postacts = []

        for i in range(self.in_dim):
            postacts_ = []
            for j in range(self.out_dim):
                xij = (
                    self.affine[j, i, 2]
                    * self.funs[j][i](
                        self.affine[j, i, 0] * x[:, [i]] + self.affine[j, i, 1]
                    )
                    + self.affine[j, i, 3]
                )
                postacts_.append(self.mask[j][i] * xij)
            postacts.append(torch.stack(postacts_))

        postacts = torch.stack(postacts)
        postacts = postacts.permute(2, 1, 0, 3)[:, :, :, 0]
        y = torch.sum(postacts, dim=2)

        return y, postacts

    def get_subset(self, in_id, out_id):
        """
        get a smaller Symbolic_KANLayer from a larger Symbolic_KANLayer (used for pruning)

        Args:
        -----
            in_id : list
                id of selected input neurons
            out_id : list
                id of selected output neurons

        Returns:
        --------
            spb : Symbolic_KANLayer

        Example
        -------
        >>> sb_large = Symbolic_KANLayer(in_dim=10, out_dim=10)
        >>> sb_small = sb_large.get_subset([0,9],[1,2,3])
        >>> sb_small.in_dim, sb_small.out_dim
        (2, 3)
        """
        sbb = Symbolic_KANLayer(self.in_dim, self.out_dim)
        sbb.in_dim = len(in_id)
        sbb.out_dim = len(out_id)
        sbb.mask.data = self.mask.data[out_id][:, in_id]
        sbb.funs = [[self.funs[j][i] for i in in_id] for j in out_id]
        sbb.funs_sympy = [
            [self.funs_sympy[j][i] for i in in_id] for j in out_id
        ]
        sbb.funs_name = [[self.funs_name[j][i] for i in in_id] for j in out_id]
        sbb.affine.data = self.affine.data[out_id][:, in_id]
        return sbb

    def fix_symbolic(
        self,
        i,
        j,
        fun_name,
        x=None,
        y=None,
        random=False,
        a_range=(-10, 10),
        b_range=(-10, 10),
        verbose=True,
    ):
        """
        fix an activation function to be symbolic

        Args:
        -----
            i : int
                the id of input neuron
            j : int
                the id of output neuron
            fun_name : str
                the name of the symbolic functions
            x : 1D array
                preactivations
            y : 1D array
                postactivations
            a_range : tuple
                sweeping range of a
            b_range : tuple
                sweeping range of a
            verbose : bool
                print more information if True

        Returns:
        --------
            r2 (coefficient of determination)

        Example 1
        ---------
        >>> # when x & y are not provided. Affine parameters are set to a = 1, b = 0, c = 1, d = 0
        >>> sb = Symbolic_KANLayer(in_dim=3, out_dim=2)
        >>> sb.fix_symbolic(2,1,'sin')
        >>> print(sb.funs_name)
        >>> print(sb.affine)
        [['', '', ''], ['', '', 'sin']]
        Parameter containing:
        tensor([[0., 0., 0., 0.],
                 [0., 0., 0., 0.],
                 [1., 0., 1., 0.]], requires_grad=True)
        Example 2
        ---------
        >>> # when x & y are provided, fit_params() is called to find the best fit coefficients
        >>> sb = Symbolic_KANLayer(in_dim=3, out_dim=2)
        >>> batch = 100
        >>> x = torch.linspace(-1,1,steps=batch)
        >>> noises = torch.normal(0,1,(batch,)) * 0.02
        >>> y = 5.0*torch.sin(3.0*x + 2.0) + 0.7 + noises
        >>> sb.fix_symbolic(2,1,'sin',x,y)
        >>> print(sb.funs_name)
        >>> print(sb.affine[1,2,:].data)
        r2 is 0.9999701976776123
        [['', '', ''], ['', '', 'sin']]
        tensor([2.9981, 1.9997, 5.0039, 0.6978])
        """
        if isinstance(fun_name, str):
            fun = SYMBOLIC_LIB[fun_name][0]
            fun_sympy = SYMBOLIC_LIB[fun_name][1]
            self.funs_sympy[j][i] = fun_sympy
            self.funs_name[j][i] = fun_name
            if x is None or y is None:
                # initialzie from just fun
                self.funs[j][i] = fun
                if not random:
                    self.affine.data[j][i] = torch.tensor([1.0, 0.0, 1.0, 0.0])
                else:
                    self.affine.data[j][i] = (
                        torch.rand(
                            4,
                        )
                        * 2
                        - 1
                    )
                return None
            else:
                # initialize from x & y and fun
                params, r2 = fit_params(
                    x,
                    y,
                    fun,
                    a_range=a_range,
                    b_range=b_range,
                    verbose=verbose,
                )
                self.funs[j][i] = fun
                self.affine.data[j][i] = params
                return r2
        else:
            # if fun_name itself is a function
            fun = fun_name
            fun_sympy = fun_name
            self.funs_sympy[j][i] = fun_sympy
            self.funs_name[j][i] = "anonymous"

            self.funs[j][i] = fun
            if not random:
                self.affine.data[j][i] = torch.tensor([1.0, 0.0, 1.0, 0.0])
            else:
                self.affine.data[j][i] = (
                    torch.rand(
                        4,
                    )
                    * 2
                    - 1
                )
            return None



"""
Full definition of a GPT Language Model, all of it in this single file.

References:
1) the official GPT-2 TensorFlow implementation released by OpenAI:
https://github.com/openai/gpt-2/blob/master/src/model.py
2) huggingface/transformers PyTorch implementation:
https://github.com/huggingface/transformers/blob/main/src/transformers\
/models/gpt2/modeling_gpt2.py
"""

import math

import torch
import torch.nn as nn
from torch.nn import functional as F

from kan_gpt.efficient_kan.model import KAN as EFFICIENT_KAN
from kan_gpt.kan.KAN import KAN as ORIGINAL_KAN
from kan_gpt.mingpt.utils import CfgNode as CN
from kan_gpt.settings import settings

# -----------------------------------------------------------------------------


class NewGELU(nn.Module):
    """
    Implementation of the GELU activation function currently in Google BERT
    repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper:
    https://arxiv.org/abs/1606.08415
    """

    def forward(self, x):
        return (
            0.5
            * x
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi)
                    * (x + 0.044715 * torch.pow(x, 3.0))
                )
            )
        )


class CausalSelfAttention(nn.Module):
    """
    A vanilla multi-head masked self-attention layer with a projection at
    the end. It is possible to use torch.nn.MultiheadAttention here but I am
    including an explicit implementation here to show that there is nothing
    too scary here.
    """

    def __init__(self, config):
        super().__init__()
        self.kan_implementation = config.kan_implementation
        KAN = self.get_KAN()

        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = KAN(width=[config.n_embd, 3 * config.n_embd])
        # output projection
        self.c_proj = KAN(width=[config.n_embd, config.n_embd])
        # regularization
        self.attn_dropout = nn.Dropout(config.attn_pdrop)
        self.resid_dropout = nn.Dropout(config.resid_pdrop)
        # causal mask to ensure that attention is only applied to the left in
        # the input sequence
        self.register_buffer(
            "bias",
            torch.tril(torch.ones(config.block_size, config.block_size)).view(
                1, 1, config.block_size, config.block_size
            ),
        )
        self.n_head = config.n_head
        self.n_embd = config.n_embd

    def get_KAN(self):
        if self.kan_implementation == "EFFICIENT_KAN":
            KAN = EFFICIENT_KAN  # type: ignore
        elif self.kan_implementation == "ORIGINAL_KAN":
            KAN = ORIGINAL_KAN  # type: ignore
        else:
            raise NotImplementedError()

        return KAN

    def forward(self, x):
        B, T, C = (
            x.size()
        )  # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head
        # forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)

        # causal self-attention;
        # Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float("-inf"))
        att = F.softmax(att, dim=-1)
        att = self.attn_dropout(att)
        y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y


class Block(nn.Module):
    """an unassuming Transformer block"""

    def __init__(self, config):
        super().__init__()
        self.kan_implementation = config.kan_implementation
        KAN = self.get_KAN()
        self.ln_1 = nn.LayerNorm(config.n_embd)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = nn.LayerNorm(config.n_embd)
        self.mlp = nn.ModuleDict(
            dict(
                c_fc=KAN(width=[config.n_embd, 4 * config.n_embd]),
                c_proj=KAN(width=[4 * config.n_embd, config.n_embd]),
                act=NewGELU(),
                dropout=nn.Dropout(config.resid_pdrop),
            )
        )
        m = self.mlp
        self.mlpf = lambda x: m.dropout(
            m.c_proj(m.act(m.c_fc(x)))
        )  # MLP forward

    def get_KAN(self):
        if self.kan_implementation == "EFFICIENT_KAN":
            KAN = EFFICIENT_KAN  # type: ignore
        elif self.kan_implementation == "ORIGINAL_KAN":
            KAN = ORIGINAL_KAN  # type: ignore

        return KAN

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlpf(self.ln_2(x))
        return x


class GPT(nn.Module):
    """GPT Language Model"""

    @staticmethod
    def get_default_config():
        C = CN()
        # either model_type or (n_layer, n_head, n_embd)
        # must be given in the config
        C.model_type = "gpt"
        C.n_layer = None
        C.n_head = None
        C.n_embd = None
        # these options must be filled in externally
        C.vocab_size = None
        C.block_size = None
        # dropout hyperparameters
        C.embd_pdrop = 0.1
        C.resid_pdrop = 0.1
        C.attn_pdrop = 0.1
        C.attn_pdrop = 0.1
        # KAN Implementation
        C.kan_implementation = settings.kan.KAN_IMPLEMENTATION
        return C

    def __init__(self, config):
        super().__init__()
        assert config.vocab_size is not None
        assert config.block_size is not None
        self.block_size = config.block_size
        self.kan_implementation = config.kan_implementation

        type_given = config.model_type is not None
        params_given = all(
            [
                config.n_layer is not None,
                config.n_head is not None,
                config.n_embd is not None,
            ]
        )
        assert type_given ^ params_given  # exactly one of these (XOR)
        if type_given:
            # translate from model_type to detailed configuration
            config.merge_from_dict(
                {
                    # names follow the huggingface naming conventions
                    # GPT-1
                    "openai-gpt": dict(
                        n_layer=12, n_head=12, n_embd=768
                    ),  # 117M params
                    # GPT-2 configs
                    "gpt2": dict(
                        n_layer=12, n_head=12, n_embd=768
                    ),  # 124M params
                    "gpt2-medium": dict(
                        n_layer=24, n_head=16, n_embd=1024
                    ),  # 350M params
                    "gpt2-large": dict(
                        n_layer=36, n_head=20, n_embd=1280
                    ),  # 774M params
                    "gpt2-xl": dict(
                        n_layer=48, n_head=25, n_embd=1600
                    ),  # 1558M params
                    # Gophers
                    "gopher-44m": dict(n_layer=8, n_head=16, n_embd=512),
                    # (there are a number more...)
                    # I made these tiny models up
                    "gpt-mini": dict(n_layer=6, n_head=6, n_embd=192),
                    "gpt-micro": dict(n_layer=4, n_head=4, n_embd=128),
                    "gpt-nano": dict(n_layer=3, n_head=3, n_embd=48),
                    "gpt-pico": dict(n_layer=1, n_head=1, n_embd=1),
                }[config.model_type]
            )

        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Embedding(config.vocab_size, config.n_embd),
                wpe=nn.Embedding(config.block_size, config.n_embd),
                drop=nn.Dropout(config.embd_pdrop),
                h=nn.ModuleList(
                    [Block(config) for _ in range(config.n_layer)]
                ),
                ln_f=nn.LayerNorm(config.n_embd),
            )
        )
        KAN = self.get_KAN()
        self.lm_head = KAN(
            width=[config.n_embd, config.vocab_size], bias_trainable=False
        )

        # init all weights, and apply a special scaled init to the residual
        # projections, per GPT-2 paper
        self.apply(self._init_weights)
        for pn, p in self.named_parameters():
            if pn.endswith("c_proj.weight"):
                torch.nn.init.normal_(
                    p, mean=0.0, std=0.02 / math.sqrt(2 * config.n_layer)
                )

        # report number of parameters (note we don't count the decoder
        # parameters in lm_head)
        n_params = sum(p.numel() for p in self.transformer.parameters())
        print("number of parameters: %.2fM" % (n_params / 1e6,))

    def get_KAN(self):
        if self.kan_implementation == "EFFICIENT_KAN":
            KAN = EFFICIENT_KAN  # type: ignore
        elif self.kan_implementation == "ORIGINAL_KAN":
            KAN = ORIGINAL_KAN  # type: ignore

        return KAN

    def kan_loss(
        self,
        x: torch.Tensor,
        lamb_l1=1.0,
        lamb_entropy=2.0,
        lamb_coef=0.0,
        lamb_coefdiff=0.0,
        small_mag_threshold=1e-16,
        small_reg_factor=1.0,
    ):

        def reg(mod):

            def nonlinear(x, th=small_mag_threshold, factor=small_reg_factor):
                return (x < th) * x * factor + (x > th) * (
                    x + (factor - 1) * th
                )

            reg_ = 0.0
            for i in range(len(mod.acts_scale)):
                vec = mod.acts_scale[i].reshape(
                    -1,
                )

                p = vec / torch.sum(vec)
                l1 = torch.sum(nonlinear(vec))
                entropy = -torch.sum(p * torch.log2(p + 1e-4))
                reg_ += (
                    lamb_l1 * l1 + lamb_entropy * entropy
                )  # both l1 and entropy

            # regularize coefficient to encourage spline to be zero
            for i in range(len(mod.act_fun)):
                coeff_l1 = torch.sum(
                    torch.mean(torch.abs(mod.act_fun[i].coef), dim=1)
                )
                coeff_diff_l1 = torch.sum(
                    torch.mean(
                        torch.abs(torch.diff(mod.act_fun[i].coef)), dim=1
                    )
                )
                reg_ += lamb_coef * coeff_l1 + lamb_coefdiff * coeff_diff_l1

            return reg_

        total_reg = torch.tensor(0.0).to(device=x.device, dtype=torch.float32)
        size = 0
        KAN = self.get_KAN()
        for mod in self.modules():
            if isinstance(mod, KAN):
                total_reg += reg(mod)
                size += 1

        mean_reg = total_reg / size
        return mean_reg

    def _init_weights(self, module):
        KAN = self.get_KAN()
        if isinstance(module, KAN):
            # torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            # if module.bias is not None:
            #     torch.nn.init.zeros_(module.bias)
            # TODO: init weights
            pass
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
        elif isinstance(module, nn.LayerNorm):
            torch.nn.init.zeros_(module.bias)
            torch.nn.init.ones_(module.weight)

    @classmethod
    def from_pretrained(cls, model_type):
        """
        Initialize a pretrained GPT model by copying over the weights
        from a huggingface/transformers checkpoint.
        """
        assert model_type in {"gpt2", "gpt2-medium", "gpt2-large", "gpt2-xl"}
        from transformers import GPT2LMHeadModel

        # create a from-scratch initialized minGPT model
        config = cls.get_default_config()
        config.model_type = model_type
        config.vocab_size = 50257  # openai's model vocabulary
        config.block_size = 1024  # openai's model block_size
        model = GPT(config)
        sd = model.state_dict()

        # init a huggingface/transformers model
        model_hf = GPT2LMHeadModel.from_pretrained(model_type)
        sd_hf = model_hf.state_dict()

        # copy while ensuring all of the parameters are aligned and
        # match in names and shapes
        keys = [
            k for k in sd_hf if not k.endswith("attn.masked_bias")
        ]  # ignore these
        transposed = [
            "attn.c_attn.weight",
            "attn.c_proj.weight",
            "mlp.c_fc.weight",
            "mlp.c_proj.weight",
        ]
        # basically the openai checkpoints use a "Conv1D" module, but we only
        # want to use a vanilla nn.Linear. this means that we have to
        # transpose these weights when we import them
        assert len(keys) == len(sd)
        for k in keys:
            if any(k.endswith(w) for w in transposed):
                # special treatment for the Conv1D weights we need to transpose
                assert sd_hf[k].shape[::-1] == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k].t())
            else:
                # vanilla copy over the other parameters
                assert sd_hf[k].shape == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k])

        return model

    def configure_optimizers(self, train_config):
        """
        This long function is unfortunately doing something very simple and
        is being very defensive: We are separating out all parameters of the
        model into two buckets: those that will experience weight decay for
        regularization and those that won't (biases, and layernorm/embedding
        weights). We are then returning the PyTorch optimizer object.
        """

        # separate out all parameters to those that will and won't experience
        # regularizing weight decay
        decay = set()
        no_decay = set()
        KAN = self.get_KAN()
        whitelist_weight_modules = (KAN,)
        blacklist_weight_modules = (torch.nn.LayerNorm, torch.nn.Embedding)
        for mn, m in self.named_modules():
            for pn, p in m.named_parameters():
                fpn = "%s.%s" % (mn, pn) if mn else pn  # full param name
                # random note: because named_modules and named_parameters
                # are recursive we will see the same tensors p many many
                # times. but doing it this way allows us to know which
                # parent module any tensor p belongs to...
                if pn.endswith("bias"):
                    # all biases will not be decayed
                    no_decay.add(fpn)
                elif isinstance(m, whitelist_weight_modules):
                    # weights of whitelist modules will be weight decayed
                    decay.add(fpn)
                elif pn.endswith("weight") and isinstance(
                    m, blacklist_weight_modules
                ):
                    # weights of blacklist modules will NOT be weight decayed
                    no_decay.add(fpn)

        # validate that we considered every parameter
        param_dict = {pn: p for pn, p in self.named_parameters()}
        inter_params = decay & no_decay
        union_params = decay | no_decay
        assert (
            len(inter_params) == 0
        ), "parameters %s made it into both decay/no_decay sets!" % (
            str(inter_params),
        )
        assert len(param_dict.keys() - union_params) == 0, (
            "parameters %s were not separated into either decay/no_decay set!"
            % (str(param_dict.keys() - union_params),)
        )

        # create the pytorch optimizer object
        optim_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(decay))],
                "weight_decay": train_config.weight_decay,
            },
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay))],
                "weight_decay": 0.0,
            },
        ]
        optimizer = torch.optim.AdamW(
            optim_groups,
            lr=train_config.learning_rate,
            betas=train_config.betas,
        )
        return optimizer

    def forward(
        self,
        idx,
        targets=None,
        lamb=0.01,
        lamb_l1=1.0,
        lamb_entropy=2.0,
        lamb_coef=0.0,
        lamb_coefdiff=0.0,
        small_mag_threshold=1e-16,
        small_reg_factor=1.0,
    ):
        device = idx.device
        b, t = idx.size()
        assert t <= self.block_size, (
            f"Cannot forward sequence of length {t}, "
            f"block size is only {self.block_size}"
        )
        pos = torch.arange(0, t, dtype=torch.long, device=device).unsqueeze(
            0
        )  # shape (1, t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(
            idx
        )  # token embeddings of shape (b, t, n_embd)
        pos_emb = self.transformer.wpe(
            pos
        )  # position embeddings of shape (1, t, n_embd)
        x = self.transformer.drop(tok_emb + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)
        logits = self.lm_head(x)

        # if we are given some desired targets also calculate the loss
        loss = None
        if targets is not None:
            loss = F.cross_entropy(
                logits.view(-1, logits.size(-1)),
                targets.view(-1),
                ignore_index=-1,
            )

            if settings.kan.KAN_IMPLEMENTATION == "ORIGINAL_KAN":
                reg = self.kan_loss(
                    x=idx,
                    lamb_l1=lamb_l1,
                    lamb_entropy=lamb_entropy,
                    lamb_coef=lamb_coef,
                    lamb_coefdiff=lamb_coefdiff,
                    small_mag_threshold=small_mag_threshold,
                    small_reg_factor=small_reg_factor,
                )
                loss = loss + lamb * reg

        return logits, loss

    @torch.no_grad()
    def generate(
        self, idx, max_new_tokens, temperature=1.0, do_sample=False, top_k=None
    ):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape
        (b,t)) and complete the sequence max_new_tokens times, feeding the
        predictions back into the model each time. Most likely you'll want
        to make sure to be in model.eval() mode of operation for this.
        """
        for _ in range(max_new_tokens):
            # if the sequence context is growing too long we must crop it at
            # block_size
            idx_cond = (
                idx
                if idx.size(1) <= self.block_size
                else idx[:, -self.block_size :]
            )
            # forward the model to get the logits for the index in the
            # sequence
            logits, _ = self(idx_cond)
            # pluck the logits at the final step and scale by desired
            # temperature
            logits = logits[:, -1, :] / temperature
            # optionally crop the logits to only the top k options
            if top_k is not None:
                v, _ = torch.topk(logits, top_k)
                logits[logits < v[:, [-1]]] = -float("Inf")
            # apply softmax to convert logits to (normalized) probabilities
            probs = F.softmax(logits, dim=-1)
            # either sample from the distribution or take the most likely
            # element
            if do_sample:
                idx_next = torch.multinomial(probs, num_samples=1)
            else:
                _, idx_next = torch.topk(probs, k=1, dim=-1)
            # append sampled index to the running sequence and continue
            idx = torch.cat((idx, idx_next), dim=1)

        return idx


if __name__ == "__main__":

    model_config = GPT.get_default_config()
    model_config.model_type = "gpt2"
    model_config.vocab_size = 5
    model_config.block_size = 10
    model = GPT(model_config)

    x = torch.zeros((1, 10), dtype=torch.long)
    y = torch.zeros((1, 10), dtype=torch.long)

    # x = x.cuda()
    # y = y.cuda()
    # model = model.cuda()

    logits, loss = model(x, y)

    print(logits.shape)
